/*
 * folder handling
 */

#include <errno.h>
#include "config.h"
#include "articles.h"
#include "news.h"
#include "term.h"
#include "menu.h"


export int  dont_sort_folders = 0;
export char *folder_directory  = NULL;

/*
 * 	file name completion and expansion
 */


expand_file_name(dest, src)
char *dest, *src;
{
    register char *cp, *dp, c;
    int parse, remap;
    char *cur_grp, *cur_art;
    
    cur_grp = current_group ? current_group->group_name : NULL;
    cur_art = (group_file_name && *group_file_name) ? group_path_name : NULL;
    
    for (dp = dest, parse = 1; c = *src; src++) {
	
	if (parse) {
	    
	    if (c == '+') {
		if (folder_directory == NULL) {
		    if (!(cp = getenv("FOLDER")))
			cp = FOLDER_DIRECTORY;
		    folder_directory = home_relative(cp);
		}
		
		cp = folder_directory;
		goto cp_str;
	    }
	    
	    if (c == '~') {
		if (src[1] != '/') {
		    msg("Can't handle ~user expansion (yet)");
		    return 0;
		}
		
		cp = home_directory;
		
	     cp_str:
		while (*cp) *dp++ = *cp++;
		if (dp[-1] != '/') *dp++ = '/';
		goto no_parse;
	    }

	    if (cur_art && c == '%' && (src[1] == ' ' || src[1] == NUL)) {
		cp = cur_art;
		while (*cp) *dp++ = *cp++;
		goto no_parse;
	    }
	    
	}

	if (c == '$' && src[1] == '(') {
	    char envar[64];
	    for (src += 2, cp = envar; (c = *src) != NUL && c != ')'; src++) 
		*cp++ = c;
	    *cp = NUL;
	    if (cp != envar) {
		if ((cp = getenv(envar)) != NULL)
		    while (*cp) *dp++ = *cp++;
		else {
		    msg("Environment variable $(%d) not set", envar);
		    return 0;
		}
	    }
	    goto no_parse;
	}
	    
	if (c == '$' && !isalnum(src[2])) {
	    remap = 0;
	    cp = NULL;
	    
	    switch (src[1]) {
	     case 'A':
		cp = cur_art;
		break;
	     case 'F':
		cp = cur_grp;
		remap = 1;
		break;
	     case 'G':
		cp = cur_grp;
		break;
	     case 'L':
		if (cp = strrchr(cur_grp, '.')) 
		    cp++;
		else 
		    cp = cur_grp;
		break;
	     case 'N':
		if (cur_art) cp = group_file_name;
		break;
	     default:
		goto copy;
	    }	    
	    src++;

	    if (!cp) {
		msg("$%c not defined on this level", c);
		return 0;
	    }

	    while (*cp) 
		if (remap && *cp == '.')
		    cp++, *dp++ = '/';
		else
		    *dp++ = *cp++;
	    goto no_parse;
	}
	
	if (c == '/')
	    if (dp != dest && dp[-1] == '/') goto no_parse;
		
     copy:
	*dp++ = c;
	parse = isspace(c);
	continue;

     no_parse:
        parse = 0;
    }

    *dp = NUL;

    return 1;
}


file_completion(path, index)
char *path;
int index;
{
    static dir_in_use = 0;
    static char *head, *tail = NULL;
    static int  tail_offset;
    
    char nbuf[FILENAME], buffer[FILENAME];
    char *dir, *base;

    if (path) {
	if (dir_in_use) {
	    close_directory();
	    dir_in_use = 0;
	}
	
	if (index < 0) return 0;

	head = path;
	tail = path + index;
    }
    
    if (!dir_in_use) {
	path = head;
	*tail = NUL;
	
	if (*path == '|') return -1;	/* no completion for pipes */
	
	if (*path == '+' || *path == '~') {
	    if (!expand_file_name(nbuf, path))
		return 0;	/* no completions */
	} else
	    strcpy(nbuf, path);
	
	if (base = strrchr(nbuf, '/')) {
	    if (base == nbuf) {
		dir = "/";
		base++;
	    } else {
		*base++ = NUL;
		dir = nbuf;
	    }
	} else {
	    base = nbuf;
	    dir = ".";
	}
	
	tail_offset = strlen(base);

	dir_in_use = list_directory(dir, base);

	return dir_in_use;
    }

    if (index)
	return help_directory();
    
    if (!next_directory(buffer)) return 0;
    
    strcpy(tail, buffer+tail_offset);
    
    return 1;
}


/*
 * 	read file names in directory 'dir' starting with 'prefix'
 *
 *	this could be speeded up by an order of magnitude by
 *	reading the directory directly into an array and sort
 *	it internally.
 *	
 */

static char dir_path[FILENAME], *dir_tail;

#ifndef HAVE_DIRECTORY

static FILE *dirf;
static int prefix_lgt;

static list_directory(dir, prefix)
char *dir, *prefix;
{
    sprintf(dir_path, "cd %s && echo %s* 2>/dev/null", dir, prefix);
    prefix_lgt = strlen(prefix);
    
    if ((dirf = popen(dir_path, "r")) == NULL) return 0;

    dir_tail = dir_path; 
    while (*dir_tail++ = *dir++); 
    dir_tail[-1] = '/';

    return 1;
}

static next_directory(buffer)
char *buffer;
{
    register char *cp;
    register int c;
    
    cp = buffer;
    while ((c = getc(dirf)) != EOF && (c != SP) && (c != NL))
	*cp++ = c;
    
    if (cp != buffer) {
	*cp = NUL;
	if (strcmp(buffer + prefix_lgt, "*")) {

	    strcpy(dir_tail, buffer);
    	    if (file_exist(dir_path, "d")) {
		*cp++ = '/';
		*cp = NUL;
	    }

	    return 1;
	}
	
    }
    
    close_directory();
    return 0;
}

help_directory()
{
    return 0;
}

static close_directory()
{
    if (dirf) {
	pclose(dirf);
	dirf = NULL;
    }
}    
#else /* HAVE_DIRECTORY */

static string_marker str_mark;
static char **completions = NULL;
static char **comp_iterator;
static char **comp_help;

/* 
 * list_directory scans the directory twice; first time to find out how
 * many matches there are, and second time to save the names, after
 * sufficient memory have been allocated to store it all.
 */

static sort_directory(f1, f2)		/* Used by qsort */
    register char **f1;
    register char **f2;
{
    return strcmp(*f1, *f2);
}

static list_directory(dir, prefix)
char *dir, *prefix;
{
    DIR *dirp;
    register Direntry *dp;
    register char *cp;
    register char **comp;
    int pflen = strlen(prefix);
    unsigned nmatch = 1;	/* No. of completions plus one */

    if ((dirp = opendir(dir)) == NULL)
	return 0;			/* tough luck */

    while ((dp = readdir(dirp)) != NULL) {
	cp = dp->d_name;
	if (*cp == '.' && (cp[1] == '\0' || (cp[1] == '.' && cp[2] == '\0')))
	    continue;
	if (strncmp(prefix, cp, pflen) == 0)
	    nmatch++;
    }
    if (nmatch == 1
	|| (completions = (char **)calloc(nmatch, sizeof(char *))) == NULL) {
	closedir(dirp);
	return 0;
    }
    mark_str(&str_mark);

    rewinddir(dirp);
    for (comp = completions; (dp = readdir(dirp)) != NULL; ) {
	cp = dp->d_name;
	if (*cp == '.' && (cp[1] == '\0' || (cp[1] == '.' && cp[2] == '\0')))
	    continue;
	if (strncmp(prefix, cp, pflen) == 0)
	    strcpy(*comp++ = alloc_str(strlen(cp)), cp);
    }
    closedir(dirp);
    *comp = (char *)0;
    qsort((char *)completions, (unsigned)(comp - completions), sizeof(char *), sort_directory);
    comp_iterator = completions;
    comp_help = completions;
    
    dir_tail = dir_path; 
    while (*dir_tail++ = *dir++); 
    dir_tail[-1] = '/';
    
    return 1;
}

static next_directory(buffer)
    register char *buffer;
{
    if (*comp_iterator != NULL) {
	strcpy(dir_tail, *comp_iterator);
	strcpy(buffer, *comp_iterator++);
	
	if ( file_exist(dir_path, "d") ) 
	    strcat(buffer, "/");
	return 1;
    }
    close_directory();
    return 0;
}

help_directory()
{
    list_completion((char *)NULL);

    if (*comp_help == NULL) comp_help = completions;
    while (*comp_help && list_completion(*comp_help))
	    comp_help++;
    
    fl;
    return 1;
}

static close_directory()
{
    if (completions) {
	release_str(&str_mark);
	free((char *)completions);
	completions = NULL;
    }
}    
#endif /* HAVE_DIRECTORY */

static int cancel_count;

fcancel(ah)
article_header *ah;
{
    if (ah->flag & A_CANCEL) {
	cancel_count--;
	ah->flag &= ~A_CANCEL;
    } else {
	cancel_count++;
	ah->flag |= A_CANCEL;
    } 
}

static folder_header()
{
    so_printxy(0, 0, "Folder: %s", current_group->group_name);

    return 1;	/* number of header lines */
}    

folder_menu(path)
char *path;
{
    FILE 			*folder;
    register article_header	*ap;
    news_header_buffer 		dgbuf;
    char 			buffer[256];
    int				more, length, re, menu_cmd, was_raw;
    memory_marker		mem_marker;
    group_header 		fake_group;
    int				cc_save;
    
    fake_group.group_name = path;
    fake_group.group_flag = G_FOLDER | G_READ;
    current_group = NULL;
    init_group(&fake_group);
    
    folder = open_file(group_path_name, OPEN_READ);
    if (folder == NULL) {
	msg("%s not found", path);
	return ME_NO_REDRAW;
    }

    was_raw = no_raw();
    s_keyboard = 0;
    
    printf("\rReading: %-.65s", path);
    clrline();
    
    current_group = &fake_group;

    mark_memory(&mem_marker);
    
    ap = alloc_art();
    
    more = 1;
    while (more && (more = get_digest_article(folder, dgbuf)) >= 0) {
	if (s_keyboard) break;
	
	ap->a_number = 0;
	ap->flag = A_FOLDER;

	ap->lines = digest.dg_lines;

	ap->hpos = digest.dg_hpos;
	ap->fpos = digest.dg_fpos;
	ap->lpos = digest.dg_lpos;

	if (digest.dg_from) {
	    length = pack_name(buffer, digest.dg_from, NAME_LENGTH);
	    ap->sender = alloc_str(length);
	    strcpy(ap->sender, buffer);
	    ap->name_length = length;
	} else {
	    ap->sender = "";
	    ap->name_length = 0;
	}
	
	if (digest.dg_subj) {
	    length = pack_subject(buffer, digest.dg_subj, &re, 255);
	    ap->replies = re;
	    ap->subject = alloc_str(length);
	    strcpy(ap->subject, buffer);
	    ap->subj_length = length;
	} else {
	    ap->replies = 0;
	    ap->subject = "";
	    ap->subj_length = 0;
	}
	
	add_article(ap);
	ap = alloc_art();
    }
    
    fclose(folder);

    if (was_raw) raw();
    
    if (s_keyboard) {
	menu_cmd = ME_NO_REDRAW;
    } else
    if (n_articles == 0) {
	msg("Not a folder (no article header)");
	menu_cmd = ME_NO_REDRAW;
    } else {
	strcpy(buffer, path);
	fake_group.group_name = buffer;	/* save for later use */
    
	if (n_articles > 1) {
	    clrdisp();
	    prompt_line = 2;
	    if (!dont_sort_folders) sort_articles();
	}
    
	cc_save = cancel_count;
	cancel_count = 0;

     reenter_menu:
	menu_cmd = menu(folder_header);

	if (cancel_count) {
	    clrdisp();
	    printf("Folder: %s\nFile:   %s\n\n", buffer, group_path_name);
	    if (cancel_count == n_articles)
		printf("Cancel all articles and remove folder? ");
	    else
		printf("Remove %d article%s from folder? ", 
		       cancel_count, cancel_count == 1 ? "" : "s");
	    fl;
	    
	    switch (yes(1)) {
	     case 1:
		printf("\n\n");
		if (cancel_count == n_articles) {
		    if (unlink(group_path_name) < 0) {
			printf("Could not unlink %s\n", group_path_name);
			sleep(3);
		    }
		} else
		    rewrite_folder();
		break;
	     case 0:
		break;
	     default:
		goto reenter_menu;
	    }
	}
	cancel_count = cc_save;
    }
    
    release_memory(&mem_marker);

    return menu_cmd;
}


rewrite_folder()
{
    register FILE *src, *dst;
    char oldfile[FILENAME], *sp;
    register int c;
    register long cnt;
    register article_header *ah, **ahp;
    register article_number n;
    
    if ((src = fopen(group_path_name, "r")) == NULL) {
	msg("Cannot open %s", group_path_name);
	return;
    }
    
    strcpy(oldfile, group_path_name);
    sp = strrchr(oldfile, '/');
    strcpy((sp == NULL ? oldfile : sp+1), "~OLD~FOLDER~");

    unlink(oldfile);
    if (link(group_path_name, oldfile) < 0) goto move_error;
    if (unlink(group_path_name) < 0) {
	if (unlink(oldfile) == 0) goto move_error;
	printf("\n\n%s was linked to %s --- cannot proceed\n",
	       group_path_name, oldfile);
	sleep(5);
	return;
    }
    
    if ((dst = fopen(group_path_name, "w")) == NULL) {
	fclose(src);
	goto move_back;
    }

    unsort_articles(1);

    printf("Compressing folder..."); fl;
    
    for (ahp = articles, n = n_articles; --n >= 0; ahp++) {
	ah = *ahp;
	if (ah->flag & A_CANCEL) continue;
	fseek(src, ah->hpos, 0);
	cnt = ah->lpos - ah->hpos;
	while (--cnt >= 0) {
	    if ((c = getc(src)) == EOF) break;
	    putc(c, dst);
	}
	putc(NL, dst);
    }
    fclose(src);
    if (ferror(dst)) {
	fclose(dst);
	goto move_back;
    }
    return;

move_back:
    if (link(oldfile, group_path_name) == 0) {
	unlink(oldfile);
	printf("Cannot create new file -- Folder restored\n");
	sleep(2);
    } else {
	printf("Cannot create new file\n\nFolder saved in %s\n",
	       oldfile);
	sleep(10);
    }
    return;
    
move_error:
    fclose(src);
    printf("\n\nCannot move folder %s to %s\n",
	   group_path_name, oldfile);
    sleep(3);
    return;
}
